"""Sanity check that ensures all colors round trip back."""
import unittest
from coloraide.css import Color


class TestRoundTrip(unittest.TestCase):
    """
    Test round trip conversions of all color spaces.

    A color that does not round trip back likely has a broken conversion.
    It may be possible that some color spaces degrade through conversion,
    but the currently supported spaces have enough integrity when performing
    a round trip to meet the required value at the default precision.
    """

    def assert_round_trip(self, color):
        """Test sRGB to HSL."""

        print('----- Color -----')
        print(color)
        c1 = Color(color)
        for space in c1.CS_MAP.keys():
            # Print the color space to easily identify which color space broke.
            print('>>> Convert to: {}'.format(space))
            c2 = c1.convert(space)
            c2.convert(c1.space(), in_place=True)
            self.assertEqual(c1.to_string(color=True), c2.to_string(color=True))

    def test_srgb(self):
        """Test sRGB."""

        self.assert_round_trip("rgb(10% 100% 50%)")

    def test_srgb_linear(self):
        """Test sRGB Linear."""

        self.assert_round_trip("color(srgb-linear 10% 100% 50%)")

    def test_hsl(self):
        """Test HSL."""

        self.assert_round_trip("hsl(128 50% 75%)")

    def test_hsv(self):
        """Test HSV."""

        self.assert_round_trip("color(hsv 300 70 50)")

    def test_hwb(self):
        """Test HWB."""

        self.assert_round_trip("hwb(90 25% 10%)")

    def test_lab(self):
        """Test Lab."""

        self.assert_round_trip("lab(50% -40 10)")

    def test_lch(self):
        """Test LCH."""

        self.assert_round_trip("lch(80% 50 270)")

    def test_xyz(self):
        """Test XYZ."""

        self.assert_round_trip("color(xyz 0.3 0.7 0.7)")

    def test_display_p3(self):
        """Test Display P3."""

        self.assert_round_trip("color(display-p3 0.3 0.4 0.9)")

    def test_rec_2020(self):
        """Test Rec 2020."""

        self.assert_round_trip("color(rec2020 0.9 0.2 0.4)")

    def test_prophoto_rgb(self):
        """Test ProPhoto RGB."""

        self.assert_round_trip("color(prophoto-rgb 0.2 0.8 0.0)")

    def test_a98_rgb(self):
        """Test A98 RGB."""

        self.assert_round_trip("color(a98-rgb 0.6 0.4 0.2)")



import math
from .. import util

G_CONST = math.pow(25, 7)
RAD2DEG = 180 / math.pi
DEG2RAD = math.pi / 180

SUPPORTED = frozenset(["76", "2000", "cmc", "94"])


def distance_euclidean(color1, color2, space="lab", **kwargs):
    """Euclidean distance."""

    lab1 = color1.convert(space)
    lab2 = color2.convert(space)

    coords1 = util.no_nan(lab1.coords())
    coords2 = util.no_nan(lab2.coords())

    total = 0
    for i, coord in enumerate(coords1):
        total += math.pow(coords2[i] - coord, 2)
    return math.sqrt(total)


def delta_e_76(color1, color2, **kwargs):
    """
    Delta E 1976 color distance formula.

    Basically this is Euclidean distance in the Lab space.
    """

    return distance_euclidean(color1, color2, space="lab")


def delta_e_94(color1, color2, kl=1, k1=0.045, k2=0.015):
    """
    Delta E 1994 color distance formula.

    http://www.brucelindbloom.com/Eqn_DeltaE_CIE94.html
    """

    lab1 = color1.convert("lab")
    lab2 = color2.convert("lab")

    l1, a1, b1 = util.no_nan(lab1.coords())
    c1 = math.sqrt(math.pow(a1, 2) + math.pow(b1, 2))
    l2, a2, b2 = util.no_nan(lab2.coords())
    c2 = math.sqrt(math.pow(a2, 2) + math.pow(b2, 2))

    dl = l1 - l2
    dc = c1 - c2

    da = a1 - a2
    db = b1 - b2

    # We never reference `dh` until the very end, and when we do, we square it
    # before using it, so we don't need the square root as described in the
    # algorithm. Instead we can just leave the result as is.
    dh = math.pow(da, 2) + math.pow(db, 2) - math.pow(dc, 2)

    sl = 1
    sc = 1 + k1 * c1
    sh = 1 + k2 * c1

    kc = 1
    kh = 1

    return math.sqrt(
        math.pow(dl / (kl * sl), 2) +
        math.pow(dc / (kc * sc), 2) +
        # Square root just the denominator as `dh` is already squared.
        dh / math.pow(kh * sh, 2)
    )


def delta_e_cmc(color1, color2, l=2, c=1):
    """
    Delta E CMC.

    http://www.brucelindbloom.com/index.html?Eqn_DeltaE_CMC.html
    """

    lab1 = color1.convert("lab")
    lab2 = color2.convert("lab")

    l1, a1, b1 = util.no_nan(lab1.coords())
    c1 = math.sqrt(math.pow(a1, 2) + math.pow(b1, 2))
    l2, a2, b2 = util.no_nan(lab2.coords())
    c2 = math.sqrt(math.pow(a2, 2) + math.pow(b2, 2))

    dl = l1 - l2
    dc = c1 - c2

    da = a1 - a2
    db = b1 - b2

    # We never reference `dh` until the very end, and when we do, we square it
    # before using it, so we don't need the square root as described in the
    # algorithm. Instead we can just leave the result as is.
    dh = math.pow(da, 2) + math.pow(db, 2) - math.pow(dc, 2)

    if l1 < 16:
        sl = 0.511
    else:
        sl = (0.040975 * l1) / (1 + 0.01765 * l1)

    sc = ((0.0638 * c1) / (1 + 0.0131 * c1)) + 0.638

    h = math.atan2(b1, a1) * RAD2DEG

    if h >= 0:
        h1 = h
    else:
        h1 = h + 360

    if 164 <= h1 <= 345:
        t = 0.56 + abs(0.2 * math.cos((h1 + 168) * DEG2RAD))
    else:
        t = 0.36 + abs(0.4 * math.cos((h1 + 35) * DEG2RAD))
    c1_4 = math.pow(c1, 4)
    f = math.sqrt(c1_4 / (c1_4 + 1900))

    sh = sc * ((f * t) + 1 - f)

    return math.sqrt(
        math.pow(dl / (l * sl), 2) +
        math.pow(dc / (c * sc), 2) +
        # Square root just the denominator as `dh` is already squared.
        dh / math.pow(sh, 2)
    )


def delta_e_2000(color1, color2, kl=1, kc=1, kh=1, **kwargs):
    """
    Calculate distance doing a direct translation of the algorithm from the CIE Delta E 2000 paper.

    TODO: this is a lot of math, we need to go through and comment this up.

    We denoted prime (L') with trailing 'p' and mean is represented with a trailing 'm'.
    Delta has a preceding 'd'. I'm not sure I was completely consistent.

    http://www2.ece.rochester.edu/~gsharma/ciede2000/ciede2000noteCRNA.pdf
    """

    lab1 = color1.convert("lab")
    lab2 = color2.convert("lab")

    l1, a1, b1 = util.no_nan(lab1.coords())
    c1 = math.sqrt(math.pow(a1, 2) + math.pow(b1, 2))
    l2, a2, b2 = util.no_nan(lab2.coords())
    c2 = math.sqrt(math.pow(a2, 2) + math.pow(b2, 2))

    cm = (c1 + c2) / 2

    c7 = math.pow(cm, 7)
    g = 0.5 * (1 - math.sqrt(c7 / (c7 + G_CONST)))

    ap1 = (1 + g) * a1
    ap2 = (1 + g) * a2

    cp1 = math.sqrt(math.pow(ap1, 2) + math.pow(b1, 2))
    cp2 = math.sqrt(math.pow(ap2, 2) + math.pow(b2, 2))

    hp1 = 0 if (ap1 == 0 and b1 == 0) else math.atan2(b1, ap1)
    hp2 = 0 if (ap2 == 0 and b2 == 0) else math.atan2(b2, ap2)

    hp1 = (hp1 + 2 * math.pi if hp1 < 0.0 else hp1) * RAD2DEG
    hp2 = (hp2 + 2 * math.pi if hp2 < 0.0 else hp2) * RAD2DEG

    dl = l2 - l1
    dc = cp2 - cp1

    hdiff = hp2 - hp1
    habs = abs(hdiff)
    if cp1 == 0.0 and cp2 == 0.0:
        dh = 0.0
    elif habs <= 180.0:
        dh = hdiff
    elif hdiff > 180.0:
        dh = hdiff - 360
    elif hdiff < -180:
        dh = hdiff + 360

    dh = 2 * math.sqrt(cp2 * cp1) * math.sin(dh * DEG2RAD / 2)

    cpm = (cp1 + cp2) / 2
    lpm = (l1 + l2) / 2

    hsum = hp1 + hp2
    if cp1 == 0 and cp2 == 0:
        hpm = hsum
    elif habs <= 180:
        hpm = hsum / 2
    elif hsum < 360:
        hpm = (hsum + 360) / 2
    else:
        hpm = (hsum - 360) / 2

    t = (
        1 -
        (0.17 * math.cos((hpm - 30) * DEG2RAD)) +
        (0.24 * math.cos(2 * hpm * DEG2RAD)) +
        (0.32 * math.cos(((3 * hpm) + 6) * DEG2RAD)) -
        (0.20 * math.cos(((4 * hpm) - 63) * DEG2RAD))
    )

    dt = 30 * math.exp(-1 * math.pow(((hpm - 275) / 25), 2))

    cpm7 = math.pow(cpm, 7)
    rc = 2 * math.sqrt(cpm7 / (cpm7 + G_CONST))
    l_temp = math.pow(lpm - 50, 2)
    sl = 1 + ((0.015 * l_temp) / math.sqrt(20 + l_temp))
    sc = 1 + 0.045 * cpm
    sh = 1 + 0.015 * cpm * t
    rt = -1 * math.sin(2 * dt * DEG2RAD) * rc

    return math.sqrt(
        math.pow((dl / (kl * sl)), 2) +
        math.pow((dc / (kc * sc)), 2) +
        math.pow((dh / (kh * sh)), 2) +
        rt * (dc / (kc * sc)) * (dh / (kh * sh))
    )


class Distance:
    """Color distancing."""

    def delta_e(self, color, method=None, **kwargs):
        """Delta E distance."""

        if method is None:
            method = self.parent.DELTA_E

        algorithm = method.lower()
        if algorithm not in SUPPORTED:
            raise ValueError("'{}' is not currently a supported distancing algorithm.".format(algorithm))

        return globals()['delta_e_{}'.format(algorithm.replace('-', '_'))](self, color, **kwargs)

    def distance(self, color, space="lab"):
        """Delta."""

        return distance_euclidean(self, color, space=space)



"""Convert utilities."""
from .. import util

WHITES = {
    "D50": [0.96422, 1.00000, 0.82521],
    "D65": [0.95047, 1.00000, 1.08883]
}


def d50_to_d65(xyz):
    """Bradford chromatic adaptation from D50 to D65."""

    m = [
        [0.9555766, -0.0230393, 0.0631636],
        [-0.0282895, 1.0099416, 0.0210077],
        [0.0122982, -0.0204830, 1.3299098]
    ]

    return util.dot(m, xyz)


def d65_to_d50(xyz):
    """
    Bradford chromatic adaptation from D65 to D50.

    The matrix below is the result of three operations:
    - convert from XYZ to retinal cone domain
    - scale components from one reference white to another
    - convert back to XYZ
    http://www.brucelindbloom.com/index.html?Eqn_ChromAdapt.html
    """

    m = [
        [1.0478112, 0.0228866, -0.0501270],
        [0.0295424, 0.9904844, -0.0170491],
        [-0.0092345, 0.0150436, 0.7521316]
    ]

    return util.dot(m, xyz)


class Convert:
    """Convert class."""

    @classmethod
    def _constrain_hue(cls, hue):
        """Constrain hue to 0 - 360."""

        return hue % 360 if not util.is_nan(hue) else hue

    @classmethod
    def _chromatic_adaption(cls, w1, w2, xyz):
        """Chromatic adaption."""

        if w1 == w2:
            return xyz
        elif w1 == WHITES["D50"] and w2 == WHITES["D65"]:
            return d50_to_d65(xyz)
        elif w1 == WHITES["D65"] and w2 == WHITES["D50"]:
            return d65_to_d50(xyz)
        else:
            raise ValueError('Unknown white point encountered: {} -> {}'.format(str(w1), str(w2)))

    def convert(self, space, *, fit=False):
        """Convert to color space."""

        space = space.lower()

        if fit:
            method = None if not isinstance(fit, str) else fit
            if not self.in_gamut(space, tolerance=0.0):
                clone = self.clone()
                result = clone.convert(space)
                result.fit(space, method=method, in_place=True)
                return result

        convert_to = '_to_{}'.format(space)
        convert_from = '_from_{}'.format(self.space())

        obj = self.parent.CS_MAP.get(space)
        if obj is None:
            raise ValueError("'{}' is not a valid color space".format(space))

        # See if there is a direct conversion route
        func = None
        coords = self._coords
        if hasattr(self, convert_to):
            func = getattr(self, convert_to)
            coords = func(coords)
        elif hasattr(obj, convert_from):
            func = getattr(obj, convert_from)
            coords = func(coords)

        # See if there is an XYZ route
        if func is None and self.space() != space:
            func = getattr(self, '_to_xyz')
            coords = func(coords)

            if space != 'xyz':
                func = getattr(obj, '_from_xyz')
                coords = func(coords)

        coords = list(coords) + [self.alpha]
        result = obj(coords)
        result.parent = self.parent

        return result

    def update(self, obj):
        """Update from color."""

        if self is obj:
            obj._coords = obj.null_adjust(obj._coords)
            return

        if not isinstance(obj, type(self)):
            obj = type(self)(obj)

        for i, value in enumerate(obj.coords()):
            self._coords[i] = value
        self.alpha = obj.alpha
        self._coords = self.null_adjust(self._coords)
        return self



"""XYZ class."""
from ._space import Space, RE_DEFAULT_MATCH
from ._gamut import GamutUnbound
from . import _parse as parse
from . import _convert as convert
import re


class XYZ(Space):
    """XYZ class."""

    SPACE = "xyz"
    DEF_VALUE = "color(xyz 0 0 0 / 1)"
    CHANNEL_NAMES = frozenset(["x", "y", "z", "alpha"])
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = convert.WHITES["D50"]

    _range = (
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([0.0, 1.0])
    )

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

        if isinstance(color, Space):
            self.x, self.y, self.z = color.convert(self.space()).coords()
            self.alpha = color.alpha
        elif isinstance(color, str):
            values = self.match(color)[0]
            if values is None:
                raise ValueError("'{}' does not appear to be a valid color".format(color))
            self.x, self.y, self.z, self.alpha = values
        elif isinstance(color, (list, tuple)):
            if not (3 <= len(color) <= 4):
                raise ValueError("A list of channel values should be of length 3 or 4.")
            self.x = color[0]
            self.y = color[1]
            self.z = color[2]
            self.alpha = 1.0 if len(color) == 3 else color[3]
        else:
            raise TypeError("Unexpected type '{}' received".format(type(color)))

    @property
    def x(self):
        """X channel."""

        return self._coords[0]

    @x.setter
    def x(self, value):
        """Shift the X."""

        self._coords[0] = self._handle_input(value)

    @property
    def y(self):
        """Y channel."""

        return self._coords[1]

    @y.setter
    def y(self, value):
        """Set Y."""

        self._coords[1] = self._handle_input(value)

    @property
    def z(self):
        """Z channel."""

        return self._coords[2]

    @z.setter
    def z(self, value):
        """Set Z channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def translate_channel(cls, channel, value):
        """Translate channel string."""

        if channel in (0, 1, 2):
            return parse.norm_float(value)
        elif channel == -1:
            return parse.norm_alpha_channel(value)
        else:
            raise ValueError("Unexpected channel index of '{}'".format(channel))



"""Cylindrical color space."""


class Cylindrical:
    """Cylindrical space."""

    def hue_name(self):
        """Hue channel name."""

        return "hue"

    def hue_index(self):
        """Get hue index."""

        return 0



"""Pro Photo RGB color class."""
from ._space import RE_DEFAULT_MATCH
from .srgb import SRGB
from . import _convert as convert
from .. import util
import re
import math


def lin_prophoto_to_xyz(rgb):
    """
    Convert an array of linear-light prophoto-rgb values to CIE XYZ using  D50.D50.

    (so no chromatic adaptation needed afterwards)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    m = [
        [0.7977604896723027, 0.13518583717574031, 0.0313493495815248],
        [0.2880711282292934, 0.7118432178101014, 0.00008565396060525902],
        [0.0, 0.0, 0.8251046025104601]
    ]

    return util.dot(m, rgb)


def xyz_to_lin_prophoto(xyz):
    """Convert XYZ to linear-light prophoto-rgb."""

    m = [
        [1.3457989731028281, -0.25558010007997534, -0.05110628506753401],
        [-0.5446224939028347, 1.5082327413132781, 0.02053603239147973],
        [0.0, 0.0, 1.2119675456389454]
    ]

    return util.dot(m, xyz)


def lin_prophoto(rgb):
    """
    Convert an array of prophoto-rgb values in the range 0.0 - 1.0 to linear light (un-corrected) form.

    Transfer curve is gamma 1.8 with a small linear portion.
    """

    et2 = 16 / 512

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i <= et2:
            result.append(i / 16)
        else:
            result.append(math.copysign(math.pow(abs_i, 1.8), i))
    return result


def gam_prophoto(rgb):
    """
    Convert an array of linear-light prophoto-rgb  in the range 0.0-1.0 to gamma corrected form.

    Transfer curve is gamma 1.8 with a small linear portion.
    """

    et = 1 / 512

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i >= et:
            result.append(math.copysign(math.pow(abs_i, 1 / 1.8), i))
        else:
            result.append(16 * i)
    return result


class ProPhoto_RGB(SRGB):
    """Pro Photo RGB class."""

    SPACE = "prophoto-rgb"
    DEF_VALUE = "color(prophoto-rgb 0 0 0 / 1)"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = convert.WHITES["D50"]

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

    @classmethod
    def _to_xyz(cls, rgb):
        """To XYZ."""

        return lin_prophoto_to_xyz(lin_prophoto(rgb))

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return gam_prophoto(xyz_to_lin_prophoto(xyz))



"""SRGB Linear color class."""
from ._space import RE_DEFAULT_MATCH
from .srgb import SRGB, lin_srgb_to_xyz, xyz_to_lin_srgb, lin_srgb, gam_srgb
from .xyz import XYZ
from . import _convert as convert
import re


class SRGB_Linear(SRGB):
    """SRGB linear."""

    SPACE = "srgb-linear"
    DEF_VALUE = "color(srgb-linear 0 0 0 / 1)"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = convert.WHITES["D65"]

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

    @classmethod
    def _to_srgb(cls, rgb):
        """Linear sRGB to sRGB."""

        return gam_srgb(rgb)

    @classmethod
    def _from_srgb(cls, rgb):
        """sRGB to linear sRGB."""

        return lin_srgb(rgb)

    @classmethod
    def _to_xyz(cls, rgb):
        """SRGB Linear to XYZ."""

        return cls._chromatic_adaption(cls.white(), XYZ.white(), lin_srgb_to_xyz(rgb))

    @classmethod
    def _from_xyz(cls, xyz):
        """XYZ to SRGB Linear."""

        return xyz_to_lin_srgb(cls._chromatic_adaption(XYZ.white(), cls.white(), xyz))



"""SRGB color class."""
from . import css_names
import re
from ...colors import srgb as generic
from ...colors import _parse as parse
from ... import util

RE_COMPRESS = re.compile(r'(?i)^#({hex})\1({hex})\2({hex})\3(?:({hex})\4)?$'.format(**parse.COLOR_PARTS))


class SRGB(generic.SRGB):
    """SRGB class."""

    DEF_VALUE = "rgb(0 0 0 / 1)"
    START = re.compile(r'(?i)\brgba?\(')
    MATCH = re.compile(
        r"""(?xi)
        (?:
            # RGB syntax
            \brgba?\(\s*
            (?:
                # Space separated format
                (?:
                    # Float form
                    (?:{float}{space}){{2}}{float} |
                    # Percent form
                    (?:{percent}{space}){{2}}{percent}
                )({slash}(?:{percent}|{float}))? |
                # Comma separated format
                (?:
                    # Float form
                    (?:{float}{comma}){{2}}{float} |
                    # Percent form
                    (?:{percent}{comma}){{2}}{percent}
                )({comma}(?:{percent}|{float}))?
            )
            \s*\) |
            # Hex syntax
            \#(?:{hex}{{6}}(?:{hex}{{2}})?|{hex}{{3}}(?:{hex})?)\b |
            # Names
            \b(?<!\#)[a-z]{{3,}}(?!\()\b
        )
        """.format(**parse.COLOR_PARTS)
    )

    HEX_MATCH = re.compile(r"(?i)#(?:({hex}{{6}})({hex}{{2}})?|({hex}{{3}})({hex})?)\b".format(**parse.COLOR_PARTS))

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

    def to_string(
        self, *, alpha=None, precision=None, fit=True, **kwargs
    ):
        """Convert to CSS."""

        if precision is None:
            precision = self.parent.PRECISION

        options = kwargs
        if options.get("color"):
            return super().to_string(alpha=alpha, precision=precision, fit=fit, **kwargs)

        # Handle hes and color names
        value = ''
        a = util.no_nan(self.alpha)
        alpha = alpha is not False and (alpha is True or a < 1.0)
        if options.get("hex") or options.get("names"):
            h = self._get_hex(options, alpha=alpha, precision=precision)
            if options.get("hex"):
                value = h
            if options.get("names"):
                length = len(h) - 1
                index = int(length / 4)
                if length in (8, 4) and h[-index:].lower() == ("f" * index):
                    h = h[:-index]
                n = css_names.hex2name(h)
                if n is not None:
                    value = n

        # Handle normal RGB function format.
        if not value:
            percent = options.get("percent", False)
            comma = options.get("comma", False)
            factor = 100.0 if percent else 255.0
            coords = util.no_nan(self.fit_coords() if fit else self.coords())

            if alpha:
                if percent:
                    template = "rgba({}%, {}%, {}%, {})" if comma else "rgb({}% {}% {}% / {})"
                else:
                    template = "rgba({}, {}, {}, {})" if comma else "rgb({} {} {} / {})"
                value = template.format(
                    util.fmt_float(coords[0] * factor, precision),
                    util.fmt_float(coords[1] * factor, precision),
                    util.fmt_float(coords[2] * factor, precision),
                    util.fmt_float(a, max(util.DEF_PREC, precision))
                )
            else:
                if percent:
                    template = "rgb({}%, {}%, {}%)" if comma else "rgb({}% {}% {}%)"
                else:
                    template = "rgb({}, {}, {})" if comma else "rgb({} {} {})"
                value = template.format(
                    util.fmt_float(coords[0] * factor, precision),
                    util.fmt_float(coords[1] * factor, precision),
                    util.fmt_float(coords[2] * factor, precision)
                )
        return value

    def _get_hex(self, options, *, alpha=False, precision=None):
        """Get the hex `RGB` value."""

        if precision is None:
            precision = self.parent.PRECISION

        hex_upper = options.get("hex_upper", False)
        compress = options.get("compress", False)
        coords = util.no_nan(self.fit_coords())

        template = "#{:02x}{:02x}{:02x}{:02x}" if alpha else "#{:02x}{:02x}{:02x}"
        if hex_upper:
            template = template.upper()

        if alpha:
            value = template.format(
                int(util.round_half_up(coords[0] * 255.0)),
                int(util.round_half_up(coords[1] * 255.0)),
                int(util.round_half_up(coords[2] * 255.0)),
                int(util.round_half_up(util.no_nan(self.alpha) * 255.0))
            )
        else:
            value = template.format(
                int(util.round_half_up(coords[0] * 255.0)),
                int(util.round_half_up(coords[1] * 255.0)),
                int(util.round_half_up(coords[2] * 255.0))
            )

        if compress:
            m = RE_COMPRESS.match(value)
            if m:
                value = m.expand(r"#\1\2\3\4") if alpha else m.expand(r"#\1\2\3")
        return value

    @classmethod
    def translate_channel(cls, channel, value):
        """Translate channel string."""

        if channel in (0, 1, 2):
            if value.startswith('#'):
                return parse.norm_hex_channel(value)
            else:
                return parse.norm_rgb_channel(value)
        elif channel == -1:
            if value.startswith('#'):
                return parse.norm_hex_channel(value)
            else:
                return parse.norm_alpha_channel(value)
        else:
            raise ValueError("Unexpected channel index of '{}'".format(channel))

    @classmethod
    def split_channels(cls, color):
        """Split channels."""

        if color[:3].lower().startswith('rgb'):
            start = 5 if color[:4].lower().startswith('rgba') else 4
            channels = []
            for i, c in enumerate(parse.RE_CHAN_SPLIT.split(color[start:-1].strip()), 0):
                if i <= 2:
                    channels.append(cls.translate_channel(i, c))
                else:
                    channels.append(cls.translate_channel(-1, c))
            if len(channels) == 3:
                channels.append(1.0)
            return cls.null_adjust(channels)
        else:
            m = cls.HEX_MATCH.match(color)
            assert(m is not None)
            if m.group(1):
                return cls.null_adjust(
                    (
                        cls.translate_channel(0, "#" + color[1:3]),
                        cls.translate_channel(1, "#" + color[3:5]),
                        cls.translate_channel(2, "#" + color[5:7]),
                        cls.translate_channel(-1, "#" + m.group(2)) if m.group(2) else 1.0
                    )
                )
            else:
                return cls.null_adjust(
                    (
                        cls.translate_channel(0, "#" + color[1] * 2),
                        cls.translate_channel(1, "#" + color[2] * 2),
                        cls.translate_channel(2, "#" + color[3] * 2),
                        cls.translate_channel(-1, "#" + m.group(4) * 2) if m.group(4) else 1.0
                    )
                )

    @classmethod
    def match(cls, string, start=0, fullmatch=True):
        """Match a CSS color string."""

        channels, end = super().match(string, start, fullmatch)
        if channels is not None:
            return channels, end
        m = cls.MATCH.match(string, start)
        if m is not None and (not fullmatch or m.end(0) == len(string)):
            if not string[start:start + 5].lower().startswith(('#', 'rgb(', 'rgba(')):
                string = css_names.name2hex(string[m.start(0):m.end(0)])
                if string is not None:
                    return cls.split_channels(string), m.end(0)
            else:
                return cls.split_channels(string[m.start(0):m.end(0)]), m.end(0)
        return None, None



"""LCH class."""
from ._space import Space, RE_DEFAULT_MATCH
from .lab import LAB
from ._cylindrical import Cylindrical
from ._gamut import GamutUnbound
from . _range import Angle, Percent
from . import _parse as parse
from . import _convert as convert
from .. import util
import re
import math

ACHROMATIC_THRESHOLD = 0.02


def lab_to_lch(lab):
    """LAB to LCH."""

    l, a, b = lab

    c = math.sqrt(math.pow(a, 2) + math.pow(b, 2))
    h = math.atan2(b, a) * 180 / math.pi

    # This is not actually part of the conversion, but is a fix-up
    # for conversion getting a bit chaotic in regards to hue when
    # chroma approaches zero. This fix-up is intended to make at
    # least colors in the sRGB range a bit more stable with conversion
    # and yield a hue of zero. This minimally affects the overall output.
    # If a 100% accurate result is desired, then we'd want to avoid doing this.
    if c < ACHROMATIC_THRESHOLD:
        h = util.NaN

    test = [l, c, LCH._constrain_hue(h)]
    return test


def lch_to_lab(lch):
    """LCH to LAB."""

    l, c, h = lch
    h = util.no_nan(h)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if c < 0.0:
        c = 0.0

    return (
        l,
        c * math.cos(h * math.pi / 180.0),
        c * math.sin(h * math.pi / 180.0)
    )


class LCH(Cylindrical, Space):
    """LCH class."""

    SPACE = "lch"
    DEF_VALUE = "color(lch 0 0 0 / 1)"
    CHANNEL_NAMES = frozenset(["lightness", "chroma", "hue", "alpha"])
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    ALPHA_COMPOSITE = "lab"
    WHITE = convert.WHITES["D50"]

    _range = (
        # I think chroma, specifically should be clamped.
        # Some libraries don't to prevent rounding issues. We should only get
        # negative chroma via direct user input, but when translating to
        # Lab, this will be corrected.
        GamutUnbound([Percent(0.0), Percent(100.0)]),
        GamutUnbound([0.0, 100.0]),
        GamutUnbound([Angle(0.0), Angle(360.0)]),
    )

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

        if isinstance(color, Space):
            self.lightness, self.chroma, self.hue = color.convert(self.space()).coords()
            self.alpha = color.alpha
        elif isinstance(color, str):
            values = self.match(color)[0]
            if values is None:
                raise ValueError("'{}' does not appear to be a valid color".format(color))
            self.lightness, self.chroma, self.hue, self.alpha = values
        elif isinstance(color, (list, tuple)):
            if not (3 <= len(color) <= 4):
                raise ValueError("A list of channel values should be of length 3 or 4.")
            self.lightness = color[0]
            self.chroma = color[1]
            self.hue = color[2]
            self.alpha = 1.0 if len(color) == 3 else color[3]
        else:
            raise TypeError("Unexpected type '{}' received".format(type(color)))

    def hue_index(self):
        """Get hue index."""

        return 2

    @property
    def lightness(self):
        """Lightness."""

        return self._coords[0]

    @lightness.setter
    def lightness(self, value):
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def chroma(self):
        """Chroma."""

        return self._coords[1]

    @chroma.setter
    def chroma(self, value):
        """chroma."""

        self._coords[1] = self._handle_input(value)

    @property
    def hue(self):
        """Hue."""

        return self._coords[2]

    @hue.setter
    def hue(self, value):
        """Shift the hue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords):
        """On color update."""

        if coords[1] < ACHROMATIC_THRESHOLD:
            coords[2] = util.NaN
        return coords

    @classmethod
    def translate_channel(cls, channel, value):
        """Translate channel string."""

        if channel in (0, 1):
            return parse.norm_float(value)
        elif channel == 2:
            return parse.norm_deg_channel(value)
        elif channel == -1:
            return parse.norm_alpha_channel(value)
        else:
            raise ValueError("Unexpected channel index of '{}'".format(channel))

    @classmethod
    def _to_lab(cls, lch):
        """To Lab."""

        return lch_to_lab(lch)

    @classmethod
    def _from_lab(cls, lab):
        """To Lab."""

        return lab_to_lch(lab)

    @classmethod
    def _to_xyz(cls, lch):
        """To XYZ."""

        return LAB._to_xyz(cls._to_lab(lch))

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return cls._from_lab(LAB._from_xyz(xyz))



"""HSL class."""
from ._space import Space, RE_DEFAULT_MATCH
from .srgb import SRGB
from ._cylindrical import Cylindrical
from ._gamut import GamutBound
from . _range import Angle, Percent
from . import _parse as parse
from . import _convert as convert
from .. import util
import re


def srgb_to_hsl(rgb):
    """SRGB to HSL."""

    r, g, b = rgb
    mx = max(rgb)
    mn = min(rgb)
    h = util.NaN
    s = 0.0
    l = (mn + mx) / 2
    c = mx - mn

    if c != 0.0:
        if mx == r:
            h = (g - b) / c
        elif mx == g:
            h = (b - r) / c + 2.0
        else:
            h = (r - g) / c + 4.0
        s = c / (1.0 - abs(2.0 * l - 1))
        h *= 60.0

    return HSL._constrain_hue(h), s * 100.0, l * 100.0


def hsl_to_srgb(hsl):
    """
    HSL to RGB.

    https://en.wikipedia.org/wiki/HSL_and_HSV#HSL_to_RGB_alternative
    """

    h, s, l = hsl
    h = util.no_nan(h)
    h = h % 360
    s /= 100.0
    l /= 100.0

    def f(n):
        """Calculate the channels."""
        k = (n + h / 30) % 12
        a = s * min(l, 1 - l)
        return l - a * max(-1, min(k - 3, 9 - k, 1))

    return f(0), f(8), f(4)


class HSL(Cylindrical, Space):
    """HSL class."""

    SPACE = "hsl"
    DEF_VALUE = "color(hsl 0 0 0 / 1)"
    CHANNEL_NAMES = frozenset(["hue", "saturation", "lightness", "alpha"])
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    ALPHA_COMPOSITE = "srgb"
    WHITE = convert.WHITES["D65"]

    _range = (
        GamutBound([Angle(0.0), Angle(360.0)]),
        GamutBound([Percent(0.0), Percent(100.0)]),
        GamutBound([Percent(0.0), Percent(100.0)])
    )

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

        if isinstance(color, Space):
            self.hue, self.saturation, self.lightness = color.convert(self.space()).coords()
            self.alpha = color.alpha
        elif isinstance(color, str):
            values = self.match(color)[0]
            if values is None:
                raise ValueError("'{}' does not appear to be a valid color".format(color))
            self.hue, self.saturation, self.lightness, self.alpha = values
        elif isinstance(color, (list, tuple)):
            if not (3 <= len(color) <= 4):
                raise ValueError("A list of channel values should be of length 3 or 4.")
            self.hue = color[0]
            self.saturation = color[1]
            self.lightness = color[2]
            self.alpha = 1.0 if len(color) == 3 else color[3]
        else:
            raise TypeError("Unexpected type '{}' received".format(type(color)))

    @property
    def hue(self):
        """Hue channel."""

        return self._coords[0]

    @hue.setter
    def hue(self, value):
        """Shift the hue."""

        self._coords[0] = self._handle_input(value)

    @property
    def saturation(self):
        """Saturation channel."""

        return self._coords[1]

    @saturation.setter
    def saturation(self, value):
        """Saturate or unsaturate the color by the given factor."""

        self._coords[1] = self._handle_input(value)

    @property
    def lightness(self):
        """Lightness channel."""

        return self._coords[2]

    @lightness.setter
    def lightness(self, value):
        """Set lightness channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords):
        """On color update."""

        if coords[1] == 0:
            coords[0] = util.NaN
        return coords

    @classmethod
    def translate_channel(cls, channel, value):
        """Translate channel string."""

        if channel == 0:
            return parse.norm_deg_channel(value)
        elif channel in (1, 2):
            return parse.norm_float(value)
        elif channel == -1:
            return parse.norm_alpha_channel(value)
        else:
            raise ValueError("Unexpected channel index of '{}'".format(channel))

    @classmethod
    def _to_srgb(cls, hsl):
        """To sRGB."""

        return hsl_to_srgb(hsl)

    @classmethod
    def _from_srgb(cls, rgb):
        """From sRGB."""

        return srgb_to_hsl(rgb)

    @classmethod
    def _to_xyz(cls, hsl):
        """To XYZ."""

        return SRGB._to_xyz(cls._to_srgb(hsl))

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return cls._from_srgb(SRGB._from_xyz(xyz))



"""Utilities."""
import decimal
import math
import numbers
import re

RE_FLOAT_TRIM = re.compile(r'^(?P<keep>-?\d+)(?P<trash>\.0+|(?P<keep2>\.\d*[1-9])0+)$')
NaN = float('nan')
INF = float('inf')
ACHROMATIC_THRESHOLD = 0.0005
DEF_PREC = 5
DEF_FIT_TOLERANCE = 0.000075
DEF_ALPHA = 1.0
DEF_MIX = 0.5
DEF_HUE_ADJ = "shorter"
DEF_DISTANCE_SPACE = "lab"
DEF_FIT = "lch-chroma"
DEF_DELTA_E = "76"


def is_number(value):
    """Check if value is a number."""

    return isinstance(value, numbers.Number)


def is_nan(value):
    """Print is "not a number"."""

    return math.isnan(value)


def no_nan(value):
    """Convert list of numbers or single number to valid numbers."""

    if is_number(value):
        return 0.0 if is_nan(value) else value
    else:
        return [(0.0 if is_nan(x) else x) for x in value]


def dot(a, b):
    """Get dot product of simple numbers, vectors, and 2D matrices and/or numbers."""

    is_a_num = is_number(a)
    is_b_num = is_number(b)
    is_a_vec = not is_a_num and is_number(a[0])
    is_b_vec = not is_b_num and is_number(b[0])
    is_a_mat = not is_a_num and not is_a_vec
    is_b_mat = not is_b_num and not is_b_vec

    if is_a_num or is_b_num:
        # Trying to dot a number with a vector or a matrix, so just multiply
        value = multiply(a, b)
    elif is_a_vec and is_b_vec:
        # Dot product of two vectors
        value = sum([x * y for x, y in zip(a, b)])
    elif is_a_mat and is_b_vec:
        # Dot product of matrix and a vector
        value = [sum([x * y for x, y in zip(row, b)]) for row in a]
    elif is_a_vec and is_b_mat:
        # Dot product of vector and a matrix
        value = [sum([x * y for x, y in zip(a, col)]) for col in zip(*b)]
    else:
        # Dot product of two matrices
        value = [[sum(x * y for x, y in zip(row, col)) for col in zip(*b)] for row in a]

    return value


def multiply(a, b):
    """Multiply simple numbers, vectors, and 2D matrices."""

    is_a_num = is_number(a)
    is_b_num = is_number(b)
    is_a_vec = not is_a_num and is_number(a[0])
    is_b_vec = not is_b_num and is_number(b[0])
    is_a_mat = not is_a_num and not is_a_vec
    is_b_mat = not is_b_num and not is_b_vec

    if is_a_num and is_b_num:
        # Multiply two numbers
        value = a * b
    elif is_a_num and not is_b_num:
        # Multiply a number and vector/matrix
        value = [multiply(a, i) for i in b]
    elif is_b_num and not is_a_num:
        # Multiply a vector/matrix and number
        value = [multiply(i, b) for i in a]
    elif is_a_vec and is_b_vec:
        # Multiply two vectors
        value = [x * y for x, y in zip(a, b)]
    elif is_a_mat and is_b_vec:
        # Multiply matrix and a vector
        value = [[x * y for x, y in zip(row, b)] for row in a]
    elif is_a_vec and is_b_mat:
        # Multiply vector and a matrix
        value = [[x * y for x, y in zip(row, a)] for row in b]
    else:
        # Multiply two matrices
        value = [[x * y for x, y in zip(ra, rb)] for ra, rb in zip(a, b)]

    return value


def divide(a, b):
    """Divide simple numbers, vectors, and 2D matrices."""

    is_a_num = isinstance(a, numbers.Number)
    is_b_num = isinstance(b, numbers.Number)
    is_a_vec = not is_a_num and isinstance(a[0], numbers.Number)
    is_b_vec = not is_b_num and isinstance(b[0], numbers.Number)
    is_a_mat = not is_a_num and not is_a_vec
    is_b_mat = not is_b_num and not is_b_vec

    if is_a_num and is_b_num:
        # Divide two numbers
        value = a / b
    elif is_a_num and not is_b_num:
        # Divide a number and vector/matrix
        value = [divide(a, i) for i in b]
    elif is_b_num and not is_a_num:
        # Divide a vector/matrix and number
        value = [divide(i, b) for i in a]
    elif is_a_vec and is_b_vec:
        # Divide two vectors
        value = [x / y for x, y in zip(a, b)]
    elif is_a_mat and is_b_vec:
        # Divide matrix and a vector
        value = [[x / y for x, y in zip(row, b)] for row in a]
    elif is_a_vec and is_b_mat:
        # Divide vector and a matrix
        value = [[x / y for x, y in zip(row, a)] for row in b]
    else:
        # Divide two matrices
        value = [[x / y for x, y in zip(ra, rb)] for ra, rb in zip(a, b)]

    return value


def cbrt(x):
    """Cube root."""

    if 0 <= x:
        return x ** (1.0 / 3.0)
    return -(-x) ** (1.0 / 3.0)


def clamp(value, mn=None, mx=None):
    """Clamp the value to the the given minimum and maximum."""

    if mn is None and mx is None:
        return value
    elif mn is None:
        return min(value, mx)
    elif mx is None:
        return max(value, mn)
    else:
        return max(min(value, mx), mn)


def adjust_precision(f, p):
    """Adjust precision and scale."""

    with decimal.localcontext() as ctx:
        if p > 0:
            # Set precision
            ctx.prec = p
        ctx.rounding = decimal.ROUND_HALF_UP

        if p == -1:
            # Full precision
            value = decimal.Decimal(f)
        elif p == 0:
            # Just round to integer
            value = decimal.Decimal(round_half_up(f))
        else:
            # Round to precision
            value = (decimal.Decimal(f) * decimal.Decimal('1.0'))
            exp = value.as_tuple().exponent
            if exp < 0 and abs(value.as_tuple().exponent) > p:
                value = value.quantize(decimal.Decimal(10) ** -p)

        if value.is_zero():
            value = abs(value)

        return float(value)


def fmt_float(f, p=0):
    """
    Set float precision and trim precision zeros.

    0: Round to whole integer
    -1: Full precision
    <positive number>: precision level
    """

    value = adjust_precision(f, p)
    string = ('{{:{}f}}'.format('.53' if p == -1 else '.' + str(p))).format(value)
    m = RE_FLOAT_TRIM.match(string)
    if m:
        string = m.group('keep')
        if m.group('keep2'):
            string += m.group('keep2')
    return string


def round_half_up(n, scale=0):
    """Round half up."""

    if scale == -1:
        return n

    mult = 10 ** scale
    return math.floor(n * mult + 0.5) / mult



"""Contrast."""


def contrast(lum1, lum2):
    """Get contrast ratio."""

    return (lum1 + 0.05) / (lum2 + 0.05) if (lum1 > lum2) else (lum2 + 0.05) / (lum1 + 0.05)


class Contrast:
    """Contrast."""

    def luminance(self):
        """Get color's luminance."""

        return self.convert("xyz").y

    def contrast(self, color):
        """Compare the contrast ration of this color and the provided color."""

        return contrast(self.luminance(), color.luminance())



"""HSV class."""
from ._space import Space, RE_DEFAULT_MATCH
from .srgb import SRGB
from .hsl import HSL
from ._cylindrical import Cylindrical
from ._gamut import GamutBound
from . _range import Angle, Percent
from . import _parse as parse
from . import _convert as convert
from .. import util
import re


def hsv_to_hsl(hsv):
    """
    HSV to HSL.

    https://en.wikipedia.org/wiki/HSL_and_HSV#Interconversion
    """

    h, s, v = hsv
    s /= 100.0
    v /= 100.0
    l = v * (1.0 - s / 2.0)

    return [
        HSV._constrain_hue(h),
        0.0 if (l == 0.0 or l == 1.0) else ((v - l) / min(l, 1.0 - l)) * 100,
        l * 100
    ]


def hsl_to_hsv(hsl):
    """
    HSL to HSV.

    https://en.wikipedia.org/wiki/HSL_and_HSV#Interconversion
    """

    h, s, l = hsl
    s /= 100.0
    l /= 100.0

    v = l + s * min(l, 1.0 - l)

    return [
        HSV._constrain_hue(h),
        0.0 if (v == 0.0) else 200.0 * (1.0 - l / v),
        100.0 * v
    ]


class HSV(Cylindrical, Space):
    """HSL class."""

    SPACE = "hsv"
    DEF_VALUE = "color(hsv 0 0 0 / 1)"
    CHANNEL_NAMES = frozenset(["hue", "saturation", "value", "alpha"])
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    GAMUT = "hsl"
    ALPHA_COMPOSITE = "srgb"
    WHITE = convert.WHITES["D65"]

    _range = (
        GamutBound([Angle(0.0), Angle(360.0)]),
        GamutBound([Percent(0.0), Percent(100.0)]),
        GamutBound([Percent(0.0), Percent(100.0)])
    )

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

        if isinstance(color, Space):
            self.hue, self.saturation, self.value = color.convert(self.space()).coords()
            self.alpha = color.alpha
        elif isinstance(color, str):
            values = self.match(color)[0]
            if values is None:
                raise ValueError("'{}' does not appear to be a valid color".format(color))
            self.hue, self.saturation, self.value, self.alpha = values
        elif isinstance(color, (list, tuple)):
            if not (3 <= len(color) <= 4):
                raise ValueError("A list of channel values should be of length 3 or 4.")
            self.hue = color[0]
            self.saturation = color[1]
            self.value = color[2]
            self.alpha = 1.0 if len(color) == 3 else color[3]
        else:
            raise TypeError("Unexpected type '{}' received".format(type(color)))

    @property
    def hue(self):
        """Hue channel."""

        return self._coords[0]

    @hue.setter
    def hue(self, value):
        """Shift the hue."""

        self._coords[0] = self._handle_input(value)

    @property
    def saturation(self):
        """Saturation channel."""

        return self._coords[1]

    @saturation.setter
    def saturation(self, value):
        """Saturate or unsaturate the color by the given factor."""

        self._coords[1] = self._handle_input(value)

    @property
    def value(self):
        """Value channel."""

        return self._coords[2]

    @value.setter
    def value(self, value):
        """Set value channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords):
        """On color update."""

        if coords[1] == 0:
            coords[0] = util.NaN
        return coords

    @classmethod
    def translate_channel(cls, channel, value):
        """Translate channel string."""

        if channel == 0:
            return parse.norm_deg_channel(value)
        elif channel in (1, 2):
            return parse.norm_float(value)
        elif channel == -1:
            return parse.norm_alpha_channel(value)
        else:
            raise ValueError("Unexpected channel index of '{}'".format(channel))

    @classmethod
    def _to_xyz(cls, hsv):
        """To XYZ."""

        return SRGB._to_xyz(cls._to_srgb(hsv))

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return cls._from_srgb(SRGB._from_xyz(xyz))

    @classmethod
    def _to_hsl(cls, hsv):
        """To HSL."""

        return hsv_to_hsl(hsv)

    @classmethod
    def _from_hsl(cls, hsl):
        """From HSL."""

        return hsl_to_hsv(hsl)

    @classmethod
    def _to_srgb(cls, hsv):
        """To sRGB."""

        return HSL._to_srgb(cls._to_hsl(hsv))

    @classmethod
    def _from_srgb(cls, rgb):
        """From sRGB."""

        return cls._from_hsl(HSL._from_srgb(rgb))



"""ColorAide Library."""
from .__meta__ import __version_info__, __version__  # noqa: F401
from .css import Color
from .colors import ColorMatch
from .util import NaN

__all__ = ("Color", "ColorMatch", "NaN")



"""Rec 2020 color class."""
from ._space import RE_DEFAULT_MATCH
from .srgb import SRGB
from .xyz import XYZ
from . import _convert as convert
from .. import util
import re
import math


def lin_2020(rgb):
    """Convert an array of rec-2020 RGB values in the range 0.0 - 1.0 to linear light (un-corrected) form."""

    alpha = 1.09929682680944
    beta = 0.018053968510807

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < beta * 4.5:
            result.append(i / 4.5)
        else:
            result.append(math.copysign(math.pow((abs_i + alpha - 1) / alpha, 1 / 0.45), i))
    return result


def gam_2020(rgb):
    """Convert an array of linear-light rec-2020 RGB  in the range 0.0-1.0 to gamma corrected form."""

    alpha = 1.09929682680944
    beta = 0.018053968510807

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i > beta:
            result.append(math.copysign(alpha * math.pow(abs_i, 0.45) - (alpha - 1), i))
        else:
            result.append(4.5 * i)
    return result


def lin_2020_to_xyz(rgb):
    """
    Convert an array of linear-light rec-2020 values to CIE XYZ using  D65.

    (no chromatic adaptation)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    m = [
        [0.6369580483012914, 0.14461690358620832, 0.1688809751641721],
        [0.2627002120112671, 0.6779980715188708, 0.05930171646986196],
        [0.000000000000000, 0.028072693049087428, 1.060985057710791]
    ]

    # 0 is actually calculated as 4.994106574466076e-17
    return util.dot(m, rgb)


def xyz_to_lin_2020(xyz):
    """Convert XYZ to linear-light rec-2020."""

    m = [
        [1.7166511879712674, -0.35567078377639233, -0.25336628137365974],
        [-0.6666843518324892, 1.6164812366349395, 0.01576854581391113],
        [0.017639857445310783, -0.042770613257808524, 0.9421031212354738]
    ]

    return util.dot(m, xyz)


class Rec2020(SRGB):
    """Rec 2020 class."""

    SPACE = "rec2020"
    DEF_VALUE = "color(rec2020 0 0 0 / 1)"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = convert.WHITES["D65"]

    def __init__(self, color=DEF_VALUE):
        """Initialize."""

        super().__init__(color)

    @classmethod
    def _to_xyz(cls, rgb):
        """To XYZ."""

        return cls._chromatic_adaption(cls.white(), XYZ.white(), lin_2020_to_xyz(lin_2020(rgb)))

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return gam_2020(xyz_to_lin_2020(cls._chromatic_adaption(XYZ.white(), cls.white(), xyz)))
